!
! This module prepares CAMS CCN for the use in mig microphysics
!
!
!
! ICON
!
! ---------------------------------------------------------------
! Copyright (C) 2004-2024, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
! Contact information: icon-model.org
!
! See AUTHORS.TXT for a list of authors
! See LICENSES/ for license information
! SPDX-License-Identifier: BSD-3-Clause
! ---------------------------------------------------------------

!----------------------------
#include "omp_definitions.inc"
!----------------------------
MODULE mo_cams_ccn   ! karoblock
  
! ICON infrastructure  
  USE mo_kind,                    ONLY: wp
  USE mo_exception,               ONLY: finish, message
  USE mo_model_domain,            ONLY: t_patch
  USE mo_grid_config,             ONLY: nroot 
  USE mo_nonhydro_types,          ONLY: t_nh_diag
  USE mo_parallel_config,         ONLY: nproma
  USE mo_loopindices,             ONLY: get_indices_c
  USE mo_impl_constants,          ONLY: min_rlcell_int, n_camsccn
  USE mo_impl_constants_grf,      ONLY: grf_bdywidth_c
  USE mo_interpolate_time,        ONLY: t_time_intp, intModeLinear
  USE mo_io_units,                ONLY: filename_max
  USE mo_fortran_tools,           ONLY: set_acc_host_or_device
  USE mo_util_string,             ONLY: int2string, associate_keyword, t_keyword_list, with_keywords
  USE mo_reader_camsccn,          ONLY: t_camsccn_reader
  USE mo_aes_phy_config,          ONLY: aes_phy_config !karoblock: should contain camsccn_filename
  USE mo_master_config,           ONLY: getModelBaseDir !karoblock
  
! External infrastruture
  USE mtime,                      ONLY: datetime, datetimeToString, MAX_DATETIME_STR_LEN !karoblock

  IMPLICIT NONE

  PRIVATE

  !> module name string
  CHARACTER(LEN=*), PARAMETER :: modname = 'mo_cams_ccn'


  PUBLIC :: cams_ccn_init
  PUBLIC :: cams_ccn_interface
  PUBLIC :: camsccn_reader
  PUBLIC :: camsccn_tintp

  TYPE(t_camsccn_reader),   ALLOCATABLE, TARGET :: camsccn_reader(:)
  TYPE(t_time_intp),        ALLOCATABLE         :: camsccn_tintp(:)


CONTAINS

  !---------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------
  !! This subroutine uploads CAMS ccn fields and updates them once a day
  SUBROUTINE cams_ccn_init(mtime_datetime, p_patch)

    TYPE(datetime), POINTER, INTENT(in) ::  mtime_datetime        !< Current datetime
    TYPE(t_patch), INTENT(in)           ::  p_patch
    
    ! Local variables
    INTEGER ::                     jg                             !< Domain index
    CHARACTER(LEN=filename_max) :: camsccn_file, camsccn_td_file  !< CAMS ccn file name
    CHARACTER(LEN=MAX_DATETIME_STR_LEN)  :: dstring

    jg = p_patch%id
 
    camsccn_file = aes_phy_config(jg)%camsccn_filename
    camsccn_td_file = generate_camsccn_filename(TRIM(camsccn_file), getModelBaseDir(), mtime_datetime)

    !WRITE(*,*) "KB: use_camsccn = ", aes_phy_config(jg)%use_camsccn
    !WRITE(*,*) "KB: camsccn_filename = ", aes_phy_config(jg)%camsccn_filename
    !CALL message ('KB: cams_ccn_init: camsccn_file = ', TRIM(camsccn_file))
    !CALL message ('KB: cams_ccn_init: camsccn_td_file = ', TRIM(camsccn_td_file))

    !WRITE(*,*) 'KB: cams_ccn_init: Call camsccn_reader'
    CALL camsccn_reader(jg)%init(p_patch, TRIM(camsccn_td_file))

    !CALL datetimeToString(mtime_datetime, dstring)
    !WRITE (*,*) "KB: cams_ccn_init: Call camsccn_tintp%init: mtime_datetime = ", TRIM(dstring)    
    CALL camsccn_tintp(jg)%init(camsccn_reader(jg), mtime_datetime, '', intModeLinear)
   

    CONTAINS

      !karoblock: generate_camsccn_filename with correct date 
      FUNCTION generate_camsccn_filename(filename_in, model_base_dir, mtime_date) RESULT(result_str)
        CHARACTER(filename_max)                 :: result_str
        CHARACTER(LEN=41), INTENT(in)           :: filename_in, &    !this is a fixed filename
             & model_base_dir                                        !path to CAMS CCN data
        TYPE(datetime), POINTER, INTENT(in)     :: mtime_date
           
        ! Local variables
        CHARACTER(LEN=MAX_DATETIME_STR_LEN) :: dtime_string
        CHARACTER(len = 30)                 :: prefix !fixed lenght: 'QUAERERE_CCNCAMS_v1_Total_CCN_'
        CHARACTER(len = 3)                  :: suffix !fixed lenght: '.nc'       
       
        prefix = filename_in(1:30)
        suffix = filename_in(39:41)

        WRITE (dtime_string,'(i4.4,2(i2.2))')                                                 &
        &                      mtime_date%date%year, mtime_date%date%month, mtime_date%date%day

        !WRITE(*,*) 'KB: model_base_dir = ',TRIM(model_base_dir)

        !camsccn_td_file = "<path>QUAERERE_CCNCAMS_v1_Total_CCN_<dtime_string>.nc"
        result_str =TRIM(model_base_dir)//prefix//TRIM(dtime_string)//suffix
        !WRITE(*,*) 'KB: generate_camsccn_filename: result_str = ',result_str
      END FUNCTION generate_camsccn_filename

  END SUBROUTINE cams_ccn_init

  !---------------------------------------------------------------------------------------
  !! This subroutine uploads CAMS ccn fields and updates them
  SUBROUTINE update_camsccn(mtime_datetime, jg, camsccn_intp)

    TYPE(datetime), POINTER, INTENT(in)  :: &
      &  mtime_datetime                    !< Current datetime
    REAL(wp), ALLOCATABLE, INTENT(inout) :: &
      &  camsccn_intp(:,:,:,:)                     !< CAMS fields taken from external file
    INTEGER, INTENT(in)                  :: &
      &  jg                                !< Domain index

    ! Local variables
    REAL(wp), ALLOCATABLE                :: &
      &  camsccncl_dat(:,:,:,:)

    ALLOCATE(camsccn_intp( nproma, camsccn_reader(jg)%nlev_cams, camsccn_reader(jg)%p_patch%nblks_c, n_camsccn+1 ))
    camsccn_intp(:,:,:,:) = 0.0_wp


   ! WRITE(*,*) 'KB: update_camsccn: Call camsccn_tintp(jg)%intp' 
    CALL camsccn_tintp(jg)%intp(mtime_datetime, camsccncl_dat)

    camsccn_intp(:,:,:,:) = camsccncl_dat(:,:,:,:)

  END SUBROUTINE update_camsccn

  !---------------------------------------------------------------------------------------
  !! Vertical interpolation of CAMS CCN
  !! This routine uses the original CAMS data which has different
  !! number of pressure levels from ICON and also different lowest/highest pressure values. 
  !! Therefore, we first move to sigma coordinates in both models. In both ICON and CAMS
  !! half level pressure are used. 
  !! The code loops on each ICON sigma level and calculates CCN hat is confined within it. 
  !! For each icon sigma layer we find lmax which is the closest
  !! CAMS level below icon_sigma2 (jk+1) and lmin which is the closest level above 
  !! icon_sigma1 (jk). There are 3 cases:
  !! CASE 1: Current ICON layer completely within one CAMS layer
  !!                       -----lmin------
  !!   --------jk-------
  !!   -------jk+1------
  !!                       -----lmax------  
  !! CASE 2: Current ICON layer covers 2 CAMS layer
  !!                       -----lmin------
  !!   --------jk-------
  !!                       ----lmin+1-----  
  !!   -------jk+1------
  !!                       -----lmax------
  !! CASE 3: Current ICON layer covers a few CAMS layer
  !!                       -----lmin------
  !!   --------jk-------
  !!                       ----lmin+1-----  
  !!                       ---- ... ------ 
  !!                       ----lmax-1----- 
  !!   -------jk+1------
  !!                       -----lmax------
  !!

  SUBROUTINE vinterp_camsccn_aerosols(i_startidx, i_endidx, nlev, pres, cams_pres_in, camsccn_intp,  camsccn )  

    INTEGER,       INTENT(in)      ::    &
      &  nlev,                           & !< ICON number of vertical levels
      &  i_startidx,                     & !< Loop indices
      &  i_endidx                          !< Loop indices
    
    REAL(wp),      INTENT(in)      ::    &
      &  camsccn_intp(:,:),              & !< time interpolated CAMS CCN from external file [m^-3]
      &  pres(:,:),                      & !< ICON diagnosed pressure
      &  cams_pres_in(:,:)                 !< CAMS half-level pressure from external file
   
    REAL(wp),      INTENT(inout)   ::    &
      &  camsccn(:,:)                      !< CAMS CCN [m^-3] on ICON vertical grid

    ! local variables
    REAL(wp)                       ::    &
      &  icon_sigma1, icon_sigma2,       & !< ICON pressure levels
      &  delta1, delta, delt1, delt,     & !< variables for looping
      &  dp_cams_surface,                & !< cams near surface pressure thickness
      &  layer_ccn                         !< ccn at specific layer
    REAL(wp) , ALLOCATABLE         ::    &
      &  cams_sigma(:,:),                & !< CAMS sigma coordinate
      &  icon_sigma(:,:)                   !< ICON sigma coordinate
    INTEGER                        ::    &
      &  jc, jk, jk1, k, lmin, lmax,     & !< Loop indices
      &  nk1, nk2, c                       !< number of vertical levels in original CAMS data

   ! REAL(wp) :: test1, test2, test3


    
    CHARACTER(len=*), PARAMETER    ::    &
      &  routine = modname//':vinterp_camsccn_aerosols'

    nk1 = size(cams_pres_in,2)
    !WRITE(*,*) 'KB: vinterp: size cams pressure levels nk1 = ', nk1
    !cams has 60 levels of base pressure, incl. surface pressure
    !cams_pres_in(jc,nk1) = surface pressure

    nk2 = size(pres,2)
    !WRITE(*,*) 'KB: vinterp: size icon pressure levels nk2 = ', nk2
    !WRITE(*,*) 'KB: vinterp: icon nlev = ', nlev
    !nk2 should be nlev+1
    !icon has 91 levels of pressure, incl. surface pressure
    !pres(jc,nlev+1) = pres(jc,nk2) = surface pressure

    !WRITE(*,*) 'KB: vinterp: allocate sigma levels'

    ALLOCATE(cams_sigma(size(cams_pres_in,1),size(cams_pres_in,2)))
    ALLOCATE(icon_sigma(size(pres,1),size(pres,2)))

    !WRITE(*,*) 'KB: vinterp:  move to sigma coordinate'

    ! move to sigma coordinate  
    DO jc = i_startidx, i_endidx  ! loop on icon horizontal index
      DO jk = 1, nk1 ! loop on icon vertical levels
          IF ( jk == 1 ) THEN
             IF( (cams_pres_in(jc,jk) .GE. 20._wp) .AND. (cams_pres_in(jc,jk) .LE. cams_pres_in(jc,nk1))) THEN
                cams_sigma(jc,jk) = cams_pres_in(jc,jk)/cams_pres_in(jc,nk1)
             ELSE
                cams_sigma(jc,jk) = 20._wp/cams_pres_in(jc,nk1)
             END IF
          ELSE
            cams_sigma(jc,jk) = cams_pres_in(jc,jk)/cams_pres_in(jc,nk1)
          END IF
      ENDDO
      DO jk = 1, nk2 
          icon_sigma(jc,jk) = pres(jc,jk)/pres(jc,nk2)
      ENDDO      
   ENDDO
   
 !        IF(jc == i_startidx) THEN
 !         !WRITE(*,*) 'KB: vinterp: cams pressure = ',cams_pres_in(jc,jk)
 !        END IF
 !       cams_sigma(jc,jk) = cams_pres_in(jc,jk)/(cams_pres_in(jc,nk1))
 !     ENDDO
 !     DO jk = 1, nk2 ! loop on icon vertical pressure ifc levels
 !        IF(jc == i_startidx) THEN
 !         !WRITE(*,*) 'KB: vinterp: icon pressure = ',pres(jc,jk)
 !        END IF
 !       icon_sigma(jc,jk) = pres(jc,jk)/pres(jc,nk2)
 !    ENDDO      
 !  ENDDO
   
   !WRITE(*,*) 'KB: before interpolating layers'
   !WRITE(*,*) 'KB: icon_sigma = ',icon_sigma(i_startidx,:),' cams_sigma = ',cams_sigma(i_startidx,:),' camsccn = ',camsccn(i_startidx,:)
       
   !WRITE(*,*) 'KB: vinterp: average over layers '

    DO jc = i_startidx, i_endidx ! loop on icon horizontal index
      DO jk = 1, nlev ! loop on icon vertical levels

        camsccn(jc,jk) = 0.0_wp
        layer_ccn = 0.0_wp

        icon_sigma1 = icon_sigma(jc,jk)
        icon_sigma2 = icon_sigma(jc,jk+1)
      
        ! ICON levels which have lower or equal pressure than CAMS lowest pressure
        IF (icon_sigma1 < cams_sigma(jc,1) .AND. icon_sigma2 <= cams_sigma(jc,1)) THEN
           layer_ccn = layer_ccn + camsccn_intp(jc,1)
        END IF

        ! ICON level partialy covers the first CAMS level
        IF (icon_sigma1 < cams_sigma(jc,1) .AND. icon_sigma2 > cams_sigma(jc,1) ) THEN
          layer_ccn = layer_ccn + camsccn_intp(jc,1) !analog to CASE 1
        END IF
        
        ! ICON levels which have higher pressure than CAMS lowest pressure
        IF (icon_sigma1 > cams_sigma(jc,1)) THEN
 
          ! loop to find lmin & lmax that are above/below jk and jk+1 respectively
          delta = 1.0_wp
          delt = 1.0_wp

          DO jk1 = 1,nk1
            delta1 = icon_sigma1 - cams_sigma(jc,jk1) 
            IF (delta1 >= 0.0_wp .AND. ABS(delta1) < delta) THEN
              lmin = jk1 
              delta = ABS(delta1)
            END IF

            delt1 = cams_sigma(jc,jk1)-icon_sigma2 
            IF (delt1 >= 0.0_wp .AND. ABS(delt1) < delt) THEN
              lmax = jk1 
              delt = ABS(delt1)
            END IF
          ENDDO

          !WRITE(*,*) 'KB: vinterp: select case'
          
          ! Current ICON layer completely within one cams layer (CASE 1)
          IF ( lmax == (lmin+1) ) THEN
          layer_ccn = layer_ccn + camsccn_intp(jc,lmax)
          !layer_ccn = layer_ccn + camsccn_intp(jc,lmin)

          ! Current ICON layer covers more than one CAMS layer (CASE 2,3)
          ELSE
            c = 0._wp       ! count number of layers
            DO k = lmin+1, lmax
              layer_ccn = layer_ccn + camsccn_intp(jc,k)
              c = c + 1._wp
            END DO
            layer_ccn = layer_ccn/c
          END IF
         
            ! this IF is for CASE 3 only
            ! IF (lmax > lmin + 2) THEN
            !  c = 0       ! count number of layers
            !  DO k = lmin+1, lmax-2
            !     layer_ccn = layer_ccn + camsccn_intp(jc,k)
            !     c = c + 1
            !  END DO
            !     layer_ccn = layer_ccn/c  !average CCN over layers (not vertically weighted)
            ! END IF 
         
            ! layer_ccn = layer_ccn + (camsccn_intp(jc,lmin) + camsccn_intp(jc,lmax-1))/2._wp          

            !END IF

        END IF
        
        !WRITE(*,*) 'KB: layer_ccn for ', jc, jk, ' = ',  layer_ccn
        IF (layer_ccn < 0.0_wp) THEN
           WRITE(*,*) "KB: layer_ccn for ", jc, jk, " = ",  layer_ccn, "s1 = ", icon_sigma1, "s2 = ", icon_sigma2
           DO k=1,nk2
              WRITE(*,*) "layer = ", k, " icon_pressure = ", pres(jc,k), "sigma =", icon_sigma(jc,k)
           ENDDO
           DO k=1,nk1
              WRITE(*,*) "layer = ", k, " cams_pressure = ", cams_pres_in(jc,k), "sigma =", cams_sigma(jc,k)
           ENDDO           
        END IF
        
        camsccn(jc,jk) = layer_ccn

        ! for checking if all gridpoints have meaningful values
        IF (camsccn(jc,jk) < 0.0_wp) THEN
          CALL finish(routine,'mo_cams_ccn: vinterp_camsccn_aerosols failed')
        END IF

      ENDDO !jk
   ENDDO !jc
   
   !WRITE(*,*) 'KB: after interpolating layers'
   !DO jk = 1, nlev
   !   test1 = cams_sigma(i_startidx,jk)/cams_sigma(i_startidx,jk)
   !   test2 = icon_sigma(i_startidx,jk)/icon_sigma(i_startidx,jk)
   !   test3 = camsccn(i_startidx,jk)/camsccn(i_startidx,jk)
      
   !   IF(test1 /= 1._wp .OR. test2 /= 1._wp .OR. test3 /= 1._wp) THEN
   !      WRITE(*,*) 'KB: istartidx = ',i_startidx,' icon_sigma = ',icon_sigma(i_startidx,jk),' cams_sigma = ',cams_sigma(i_startidx,jk),' camsccn = ',camsccn(i_startidx,jk),' pres_base = ',cams_pres_in(i_startidx,jk)
   !   END IF
   !END DO

    DEALLOCATE(cams_sigma)
    DEALLOCATE(icon_sigma)

  END SUBROUTINE vinterp_camsccn_aerosols



  !---------------------------------------------------------------------------------------
  !! This subroutine calls CCN time updates and vertical interpolation routines
  !! and updates the diagnostic ccn pointer
  SUBROUTINE cams_ccn_interface(mtime_datetime, pt_patch, pt_diag, lacc)

    CHARACTER(len=*), PARAMETER :: &
      &  routine = modname//':cams_ccn_interface'

    TYPE(datetime), POINTER, INTENT(in) :: &
      &  mtime_datetime          !< Current datetime
    TYPE(t_patch), TARGET, INTENT(in) :: &
      &  pt_patch                !< Grid/patch info
    TYPE(t_nh_diag), TARGET, INTENT(inout) :: &
      &  pt_diag                 !< the diagnostic variables

    LOGICAL, OPTIONAL, INTENT(in) :: lacc ! If true, use openacc

    ! Local variables
  
    REAL(wp),  ALLOCATABLE ::  &
      &  camsccn_intp(:,:,:,:)      !< time interpolated CAMS ccn fields taken from external file 

    INTEGER ::                 &
      &  jk, jc, jb, jt,       &
      &  rl_start, rl_end,     &
      &  i_startblk, i_endblk, &
      &  i_startidx, i_endidx

    LOGICAL :: lzacc

    CALL set_acc_host_or_device(lzacc, lacc)

  
        rl_start   = grf_bdywidth_c+1
        rl_end     = min_rlcell_int
        i_startblk = pt_patch%cells%start_block(rl_start)
        i_endblk   = pt_patch%cells%end_block(rl_end)

        ! Compatibility checks still missing
        !WRITE(*,*) 'KB: cams_ccn_interface: call update_camsccn' 
        CALL update_camsccn(mtime_datetime, pt_patch%id, camsccn_intp)

        DO jt=1, n_camsccn
!$OMP PARALLEL
!$OMP DO PRIVATE(jb,i_startidx,i_endidx) ICON_OMP_DEFAULT_SCHEDULE
          DO jb = i_startblk,i_endblk
            CALL get_indices_c(pt_patch,jb,i_startblk,i_endblk,i_startidx,i_endidx,rl_start,rl_end)
            IF (i_startidx>i_endidx) CYCLE
            pt_diag%camsccn(:,:,jb,jt) = 0._wp !karoblock
            
            !WRITE(*,*) 'KB: cams_ccn_interface: call vinterp_camsccn_aerosols'
            CALL vinterp_camsccn_aerosols(i_startidx, i_endidx, nlev=pt_patch%nlev,                                 &
                 &                   pres=pt_diag%pres_ifc(:,:,jb), cams_pres_in=camsccn_intp(:,:,jb,n_camsccn+1),  &
                 &                   camsccn_intp=camsccn_intp(:,:,jb,jt), camsccn=pt_diag%camsccn(:,:,jb,jt))


         END DO
!$OMP END DO NOWAIT
!$OMP END PARALLEL
        END DO ! jt supersaturation loop

      END SUBROUTINE cams_ccn_interface

      
END MODULE mo_cams_ccn
