! ICON
!
! ---------------------------------------------------------------
! Copyright (C) 2004-2024, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
! Contact information: icon-model.org
!
! See AUTHORS.TXT for a list of authors
! See LICENSES/ for license information
! SPDX-License-Identifier: BSD-3-Clause
! ---------------------------------------------------------------

MODULE mo_reader_camsccn   !karoblock

  USE mo_kind,                    ONLY: dp, wp, i8
  USE mo_parallel_config,         ONLY: get_nproma
  USE mo_exception,               ONLY: finish
  USE mo_reader_abstract,         ONLY: t_abstract_reader
  USE mo_impl_constants,          ONLY: n_camsccn
  USE mo_io_units,                ONLY: FILENAME_MAX
  USE mo_model_domain,            ONLY: t_patch
  USE mo_netcdf_errhandler,       ONLY: nf
  USE mo_netcdf
  USE mtime,                      ONLY: julianday, juliandelta, getJulianDayFromDatetime,   &
                                    &   datetime, newdatetime, deallocatedatetime,          &
                                    &   OPERATOR(+), ASSIGNMENT(=),                         &
                                    &   no_of_ms_in_a_day, no_of_ms_in_a_hour,              &
                                    &   no_of_ms_in_a_minute, no_of_ms_in_a_second
  USE mo_mpi,                     ONLY: my_process_is_mpi_workroot, process_mpi_root_id, p_comm_work, p_bcast
  USE mo_read_netcdf_distributed, ONLY: distrib_nf_open, distrib_read, distrib_nf_close, idx_blk_time
  USE mo_fortran_tools,           ONLY: t_ptr_4d

  IMPLICIT NONE

  PRIVATE

  PUBLIC :: t_camsccn_reader

  TYPE, EXTENDS(t_abstract_reader) :: t_camsccn_reader

    TYPE(t_patch), POINTER      :: p_patch => NULL()
    CHARACTER(len=NF_MAX_NAME)  :: varnames(n_camsccn+1)    
    CHARACTER(len=FILENAME_MAX) :: filename
    INTEGER                     :: fileid, dist_fileid, nlev_cams
    LOGICAL                     :: lopened = .FALSE.

  CONTAINS

    PROCEDURE :: init            => camsccn_init_reader
    PROCEDURE :: get_one_timelev => camsccn_get_one_timelevel
    PROCEDURE :: get_times       => camsccn_get_times
    PROCEDURE :: deinit          => camsccn_deinit_reader

    PROCEDURE :: get_nblks       => camsccn_get_nblks
    PROCEDURE :: get_npromz      => camsccn_get_npromz

  END TYPE t_camsccn_reader

  CHARACTER(len=*), PARAMETER :: modname = 'mo_reader_camsccn'

CONTAINS

  SUBROUTINE camsccn_init_reader(this, p_patch, filename)
    CLASS(t_camsccn_reader),    INTENT(inout) :: this
    TYPE(t_patch),      TARGET, INTENT(in   ) :: p_patch
    CHARACTER(len=*),           INTENT(in   ) :: filename

    CHARACTER(len=*), PARAMETER :: routine = 'camsccn_init_reader'

    this%filename = TRIM(filename)
    !WRITE(*,*) 'KB: camsccn_init_reader: this%filename=', this%filename
    
    this%varnames(1)  = "ccn_ss02"		  ! CCN number conc. at Ssat=0.2% (m-3)
    this%varnames(2)  = "ccn_ss04"		  ! CCN number conc. at Ssat=0.4% (m-3)
    this%varnames(3)  = "ccn_ss06"		  ! CCN number conc. at Ssat=0.6% (m-3)
    this%varnames(4)  = "ccn_ss08"		  ! CCN number conc. at Ssat=0.8% (m-3)
    this%varnames(5)  = "ccn_ss10"		  ! CCN number conc. at Ssat=1.0% (m-3)
    this%varnames(6)  = "half_level_pressure_base"     ! Pressure at base of layer (Pa)
	  
	  
    this%nlev_cams = 60
    this%p_patch => p_patch
    
    !WRITE(*,*) 'KB: camsccn_init_reader: this%varnames(1)=', this%varnames(1)
    !WRITE(*,*) 'KB: camsccn_init_reader: this%varnames(6)=', this%varnames(7)
    !WRITE(*,*) 'KB: camsccn_init_reader: this%nlev_cams=', this%nlev_cams
    
    
    IF (.NOT. this%lopened) THEN
       IF (my_process_is_mpi_workroot()) THEN
        !WRITE(*,*) 'KB: camsccn_init_reader: call nf_open(this%filename)'
        CALL nf(nf_open(this%filename, nf_nowrite, this%fileid), routine)
      ENDIF
      this%dist_fileid = distrib_nf_open(TRIM(this%filename))
      this%lopened = .TRUE.
    ENDIF

  END SUBROUTINE camsccn_init_reader

  SUBROUTINE camsccn_get_times (this, times)

    CLASS(t_camsccn_reader), INTENT(inout) :: this
    TYPE(julianday), ALLOCATABLE, INTENT(out) :: times(:)
   
    INTEGER                       :: tvid, tdid
    CHARACTER(len=NF_MAX_NAME)    :: cf_timeaxis_string
    CHARACTER(len=:), ALLOCATABLE :: epoch
    CHARACTER(len=:), ALLOCATABLE :: base_timeaxis_unit
    TYPE(datetime), POINTER       :: epoch_datetime
    TYPE(julianday)               :: epoch_jd
    TYPE(juliandelta)             :: offset
    INTEGER(i8)                   :: time_multiplicator
    REAL(wp), ALLOCATABLE         :: times_read(:)
    INTEGER                       :: ntimes
    INTEGER                       :: i

    CHARACTER(len=*), PARAMETER :: routine = 'camsccn_get_times'

    IF (my_process_is_mpi_workroot()) THEN
      !WRITE(*,*) 'KB: camsccn_get_times: call nf(time)'
       
      CALL nf(nf_inq_varid(this%fileid, "time", tvid), routine)
      CALL nf(nf_inq_dimid(this%fileid, "time", tdid), routine)
      CALL nf(nf_inq_dimlen(this%fileid, tdid, ntimes), routine)      

      ALLOCATE(times_read(ntimes))

      CALL nf(nf_get_var(this%fileid, tvid, times_read), routine)
      CALL nf(nf_get_att(this%fileid,   tvid, "units", cf_timeaxis_string), routine)

    ENDIF

    CALL p_bcast(ntimes, process_mpi_root_id, p_comm_work)
    IF (.NOT. ALLOCATED(times_read)) THEN
      ALLOCATE(times_read(ntimes))
    ENDIF
    CALL p_bcast(times_read, process_mpi_root_id, p_comm_work)
    CALL p_bcast(cf_timeaxis_string, process_mpi_root_id, p_comm_work)

    CALL get_cf_timeaxis_desc(TRIM(cf_timeaxis_string), epoch, base_timeaxis_unit)

    epoch_datetime => newdatetime(epoch)
    CALL getJulianDayFromDatetime(epoch_datetime, epoch_jd)
    CALL deallocateDatetime(epoch_datetime)
    
    SELECT CASE (base_timeaxis_unit)
    CASE('days')
      time_multiplicator = no_of_ms_in_a_day
    CASE('hours')
      time_multiplicator = no_of_ms_in_a_hour
    CASE('minutes')
      time_multiplicator = no_of_ms_in_a_minute
    CASE('seconds')
      time_multiplicator = no_of_ms_in_a_second
    END SELECT
    
    ALLOCATE(times(ntimes))

    DO i = 1, ntimes
      offset%sign = '+'
      offset%day  = INT((time_multiplicator * times_read(i))/86400000.0_wp,i8)
      offset%ms   = NINT(MOD(time_multiplicator * times_read(i), 86400000.0_wp),i8)
      times(i) = epoch_jd + offset
    ENDDO
    
  END SUBROUTINE camsccn_get_times


  SUBROUTINE camsccn_get_one_timelevel(this, timelevel, varname, dat)
    CLASS(t_camsccn_reader), INTENT(inout)  :: this
    INTEGER, INTENT(in   )               :: timelevel
    CHARACTER(len=*), INTENT(in   )      :: varname
    REAL(wp), ALLOCATABLE, INTENT(inout) :: dat(:,:,:,:)
    REAL(wp), ALLOCATABLE, TARGET        :: temp(:,:,:,:)
    TYPE(t_ptr_4d)                       :: tmp(1)
    INTEGER                              :: var_dimlen(3),var_start(3), var_end(3),jt


    ALLOCATE(temp(get_nproma(), this%nlev_cams, this%p_patch%nblks_c, n_camsccn+1))
    IF (ALLOCATED(dat)) DEALLOCATE(dat)
    ALLOCATE( dat(get_nproma(), this%nlev_cams, this%p_patch%nblks_c, n_camsccn+1))

      var_dimlen(2) = SIZE(temp, 2) ! number of vertical levels
      var_dimlen(3) = 1             ! number of time steps = 1
      var_start(:)  = (/1, 1, 1/)
      var_end(:)    = var_dimlen(:)
      var_start(3)  = timelevel
      var_end(3)    = timelevel

    temp(:,:,:,:) = -1.0_wp

    IF (.NOT. this%lopened) &
      CALL finish(modname, 'CAMS CCN file not open!')
    
    IF (TRIM(varname) /= '') &
      CALL finish(modname, 'CAMS CCN: Variables not read!')

    tmp(1)%p => temp

    DO jt = 1, n_camsccn+1
       !WRITE(*,*) 'KB: camsccn_get_one_timelevel: call distrib_read'
       
       CALL distrib_read(this%dist_fileid, this%varnames(jt), tmp, &
         & (/this%p_patch%cells%dist_io_data/), edim=var_dimlen(2:3), dimo=idx_blk_time, &
         & start_ext_dim=var_start(2:3), end_ext_dim=var_end(2:3))

         dat(:,:,:,jt) = temp(:,:,:,1)
    ENDDO

    DEALLOCATE(temp)
    
  END SUBROUTINE camsccn_get_one_timelevel

  FUNCTION camsccn_get_nblks (this) RESULT(nblks)
    CLASS(t_camsccn_reader), INTENT(in   ) :: this
    INTEGER                                :: nblks
    nblks = this%p_patch%nblks_c
  END FUNCTION camsccn_get_nblks

  FUNCTION camsccn_get_npromz (this) RESULT(npromz)
    CLASS(t_camsccn_reader), INTENT(in   ) :: this
    INTEGER                                :: npromz
    npromz = this%p_patch%npromz_c
  END FUNCTION camsccn_get_npromz

  SUBROUTINE camsccn_deinit_reader(this)
    CLASS(t_camsccn_reader), INTENT(inout) :: this
    CHARACTER(len=*), PARAMETER :: routine = 'camsccn_deinit_reader'
    IF (ASSOCIATED(this%p_patch)) NULLIFY(this%p_patch)
    IF (this%lopened) THEN
      IF (my_process_is_mpi_workroot()) THEN
        CALL nf(nf_close(this%fileid), routine)
      END IF
      CALL distrib_nf_close(this%dist_fileid)
    END IF
  END SUBROUTINE camsccn_deinit_reader

  !karoblock: copied from mo_reader_sst_sic
  SUBROUTINE get_cf_timeaxis_desc(cf_timeaxis_string, epoch, base_timeaxis_unit)
    CHARACTER(len=*), INTENT(in) :: cf_timeaxis_string
    CHARACTER(len=:), ALLOCATABLE, INTENT(out) :: epoch, base_timeaxis_unit
    
    ! The CF convention allows for a timezone to be included. We will
    ! ignore that one for all , but gets stored to word(5), if
    ! provided, to keep the algorithm simple.

    CHARACTER(len=16) :: word(5)
    INTEGER :: pos1, pos2, n

    pos1 = 1; pos2 = 0; n = 0;
    word(:) = ""

    DO
      pos2 = INDEX(cf_timeaxis_string(pos1:), " ")
      IF (pos2 == 0) THEN
        n = n + 1
        word(n) = cf_timeaxis_string(pos1:)
        EXIT
      ENDIF
      n = n + 1
      word(n) = cf_timeaxis_string(pos1:pos1+pos2-2)
      pos1 = pos2+pos1
    ENDDO

 ! correct the date part
    normalize_date: BLOCK
      INTEGER :: idx1, idx2
      INTEGER :: year, month, day
      idx1 = INDEX(word(3), '-')    
      idx2 = INDEX(word(3)(idx1+1:), '-')+idx1
      READ(word(3)(      :idx1-1),*) year
      READ(word(3)(idx1+1:idx2-1),*) month
      READ(word(3)(idx2+1:      ),*) day
      WRITE(word(3),'(i0,a,i2.2,a,i2.2)') year, '-', month, '-', day
    END BLOCK normalize_date
    
    IF (word(4) /= "") THEN
      epoch = TRIM(word(3))//'T'//TRIM(word(4))
    ELSE
      epoch = TRIM(word(3))
    ENDIF

    base_timeaxis_unit = TRIM(word(1))

  END SUBROUTINE get_cf_timeaxis_desc
  
END MODULE mo_reader_camsccn
